((python-mode . ((indent-tabs-mode . nil)))
 (autoconf-mode . ((indent-tabs-mode . t)))
 (c-mode . ((c-file-style . "stroustrup")
	    (indent-tabs-mode . t)
	    (tab-width . 8)
	    (c-basic-offset . 8)
	    (c-file-offsets .
			    ((block-close . 0)
			     (brace-list-close . 0)
			     (brace-list-entry . 0)
			     (brace-list-intro . +)
			     (case-label . 0)
			     (class-close . 0)
			     (defun-block-intro . +)
			     (defun-close . 0)
			     (defun-open . 0)
			     (else-clause . 0)
			     (inclass . +)
			     (label . 0)
			     (statement . 0)
			     (statement-block-intro . +)
			     (statement-case-intro . +)
			     (statement-cont . +)
			     (substatement . +)
			     (topmost-intro . 0))))))
